# Beginning Blender Python.py:
#   Feb 14 2024
#
# Written by Terry Erickson
# Written for and tested using Blender v4.02
#
#
#
#
#
#
#
#
# This is a basic overview of how to use the Blender
# Python API to work with Blender via scripting. It
# is not meant to be a complete reference, but rather
# a quick start guide for beginners. I have written
# it to be a document like I wish I had when I was
# starting to program Blender with Python code.
#
# There is only minimal 'error protection' used in
# this script as it is meant to be only an example of
# basic tasks. Users are encouraged to create code to
# check if required data (or Objects) are present
# before involking operations in their scripts.
#
# This is a compliment file to the included
# Beginning_Blender_Python.blend file, which already
# has many Objects and collections which will be
# used in this script. If this script is not already
# added to the .blend file, it can be 'open'ed in
# the "Text Editor" window of the .blend file.
#
# Running this script will execute ALL of the code
# which is not commented out. So it will be difficult
# to witness the results of each section as it is
# processed. Therefore, feel free to comment out any
# portion(s) desired before running this script.
# Optionally, you can break the script into sections
# and run each individually.
#
#
#
#
# Blender provides help!
#
# Not only does Blender provide an API manual, it
# also allows us to get specific references from
# many menus and input areas. If we have Python
# tooltips enabled (setting is in: Edit, Preferences,
# Interface, Tooltips, Python tooltips), we can
# hover over an input value to see the reference
# as a tooltip. we can also copy it to the
# clipboard using any of the following keystrokes:
#   CTRL+C            Copy the value only
#   SHIFT+CTRL+C      Copy the data path
#   SHIFT+CTRL+ALT+C  Copy the full data path
#
# While learning how things work, we can play with
# commands using a "Python Console" area in our
# Blender project. This allows us to start typing
# a line of code and get guidance about what may
# be expected in the line of code. Consult the
# Blender manual for more information.
#
#
#
#
#
#
#
#
# INDEX OF SECTIONS:
#
#    LINE   SECTION
#   ------  ----------------------------------------
#      110  Importing modules/libraries:
#      166  Using path shortcuts:
#      217  Avoiding common problems:
#      330  Definitions:
#      451  Lists:
#      581  Function summary:
#      692  Functions:
#     1324  Load Object references into a list:
#     1438  Renaming:
#     1456  Transforming:
#     1513  "Apply" the transformation:
#     1581  Finding Objects:
#     1732  Moving the 3D cursor:
#     1784  Changing Object Origins:
#     1839  Creating an Empty:
#     1895  Parenting Objects:
#     1955  Joining Objects:
#     2063  Use a Modifier:
#     2143  Use a Constraint:
#     2260  Mess with the Mesh:
#     2384  Lets text:
#     2451  Create a mesh Object from raw data:
#     2641  Make a pyramid of Pyramids:
#     3125  Keyframes and animation:
#     3526  Closing:
#
#
#
#
#
#
#
#
# Importing modules/libraries:
#
# Usually the first thing we want to do is import any
# modules needed. Doing this first insures we have
# the ability to use those operations in our script.
#
# We can either load entire or specific libraries.
# Once familiar with which libraries we need, loading
# specific ones uses less resources and may improve the
# performance of our script run.
#
#
# Modules
#
#
# The BPY library is the main Blender data interface.
# Here we are importing the entire library.
#
# Interface
import bpy 
#
#
# Bmesh is Blenders internal mesh editing API.
#
# Bmesh
import bmesh
#
#
# The Python MATH library contains support for many
# useful math operations and definitions. We can
# import the entire library:
#
# Math
#import math
#
# Or import specific parts of it:
#
# Random
#import random
#
# Radians
#from math import radians
#
# Mathutils
#import mathutils
#
# Vector
#from mathutils import Vector
#
#
#
#
#
#
#
#
# Using path shortcuts:
#
# Many programmers like to use 'shortcuts' to replace
# long names with abbreviated variable names. This can
# make code lines shorter, but may impede a beginner
# with less understandable code. I suggest beginners
# start with using full reference names until they
# are at least comfortable with how things are done.
#
# Below is a sample of how some shortcuts could be
# done. We can use these with whatever variable
# names we desire. Doing this right after importing
# libraries gives us a way to look up our shortcuts
# in case we forget them.
#
# I do not use many shortcuts in this document so
# that beginners may learn what paths match with
# tasks. Some shortcuts are used though to make
# very lengthy lines more manageable.
#
#
# Common shortcuts
#
# BPY
#B = bpy
#
# Context
#BC = bpy.context
#
# Scene objects
#BCSO = bpy.context.scene.objects
#
# View Layer
#BCV = bpy.context.view_layer
#
# Data
#BD = bpy.data
#
# Collection
#BDC = bpy.data.collections
#
# Operation
#BO = bpy.ops
#
#
#
#
#
#
#
#
# Avoiding common problems:
#
# Before we jump into the heart of the scripts, it is
# important to cover a few things that often cause
# errors for beginners.
#
#
# 
# Meter
# 
# The Blender API uses METRIC values! All length and
# distance values need to be in METERS. Even if our
# "Scene" properties "Unit System" is set to 'Imperial'
# (inches or feet), values entered through Python must
# be in Meters to produce desired results.
#
# It is easy to convert to Meters using:
#   Meter = Inches / 39.37
# For example:
#   INCHES = FEET = PYTHON VALUE
#   ------   ----   ------------
#     0.01          0.000254
#      0.1          0.00254
#        1          0.0254
#       10          0.254
#       12      1   0.3048
#       36      3   0.9144
#    39.37          1
#      120     10   3.048
#
# This script includes a conversion functions that
# can be used if desired.
#
#
#
# Radian
#
# Although Blender allows angle values to be entered
# in degrees through its slider inputs, values in
# Python code need to be in Radians.
#
# It is easy to convert to Radians using:
#   Radian = Degree * PI/180
#          = Degree * 0.017453293
#
# This script includes a conversion functions that
# can be used if desired.
#
#
#
# Indentation
#
# Those not familiar with the Python scripting language
# would be strongly advised to learn at least basic
# Python coding before attempting to code for their
# Blender project. Many programming languages have
# different keywords or styles to be aware of.
#
# Python demands well formatted (indented) code. It
# uses indentation to help define subsections of
# code, such as loop bodies or multi-line statements.
# I use a '4 space' indent, like:
#onl = [
#    "Cone",
#    "Cube",
#    "Cylinder",
#    "Plane",
#    "Sphere",
#    "Torus"
#    ]
#
# I do this in many other languages also, even though
# it is for personal clarity rather than required.
#
#
#
# Comments
#
# Comment lines (non-executable content) always start
# with a "#" (numeric sign). In addition to the main
# comments throughout this document, I include many
# lines of actual code which are commented out as
# they are optional. Comment lines must follow the
# correct indentation rules also.
#
# When I am editing code, I often find it to be very
# useful to copy a line of code before I modify it
# and comment out the original line. Such as:
#   # XYZ
#   #obj.location = (.7, .35, 0)   <-- Original values
#   obj.location = (.9, .27, 0)    <-- New values
# This gives me a reference to the original values
# in case my newly calculated values are incorrect.
# 
# As a personal preference, I like to add space
# (mainly empty lines) to my code. I think this makes
# it more readable (instead of just a huge blob of
# code), but also gives me room for comments before
# each statement. This often is very helpful if I
# need to modify a script years later, as comments
# tell me why I did certain things in certain ways
# instead of having to figure it out by searching
# through the rest of the code.
#
# Feel free to use commenting however desired!
#
#
#
#
#
#
#
#
# Definitions:
#
# It is not required to pre-define variables in Python!
# But I like to do it for three reasons:
#   It allows me to easily keep track of names I use,
#     which helps avoid duplicate names
#   It allows me to set default values
#   It gives me a quick reference when writing code
#
# For this document, I use temporary variables to hold
# values after they are obtained from one source until
# they are transferred somewhere else. This helps to
# show individual actions on a line instead of having
# multiple actions which may confuse beginners.
#
#
# Location
#
# X
x = 0
#
# Y
y = 0
#
# Z
z = 0
#
#
# Rotation
#
# X
xr = 0
#
# Y
yr = 0
#
# Z
zr = 0
#
# Angle
#
# 1 radian
r1 = 0.017453293
#
# Angle in radians
air = 0
#
# Angle in degrees
aid = 0
#
#
# Scale
#
# X
xs = 0
#
# Y
ys = 0
#
# Z
zs = 0
#
#
# Dimension
#
# X
xd = 0
#
# Y
yd = 0
#
# Z
zd = 0
#
#
# 3D Cursor
#
# Loc
c3dl = (0, 0, 0)
#
# Rot
c3dr = (0, 0, 0)
#
# Scale
c3ds = (0, 0, 0)
#
#
# Loop
#
# General
l = 0
#
# Object
lo = 0
#
# Row
lr = 0
#
# Column
lc = 0
#
# Range
ln = 0
#
#
# Random result
rn = 0
#
#
# Flag
#
# FLG
flg = False
#
#
#
#
#
#
#
#
# Lists:
#
# Many programming languages use ARRAYS to store data.
# Python uses LISTS which work in a similar manner.
# Lists allow us to get data using an 'index' value
# (either individual or through a loop). The first
# index is zero (0).
#
# For my Blender scripts, I like to store a reference
# to all of the Objects I will use in a list. This
# lets me access any Object reference with a single
# variable name and index, such as:
#   o[0].location = .054
# where o[3] would be my "Color_ball" object. I always
# have a 'name' and 'reference' list. These let me
# build the Object reference list via a loop and
# also provide me with a way to look up what index
# any of my Objects are at. The order does not matter,
# but alphabetic order is easier to look up names.
#
# Doing this is Optional, but I use this method in
# this tutorial to avoid having numerous Object name
# variables or a single variable (like OBJ) used to
# reference different Objects (especially since some
# sections work with multiple Objects at once).
#
#
# Lists
#
#
# Object names
# ONL
onl = [
    "Color_ball",
    "Cone",
    "Cube",
    "Cylinder",
    "Plane",
    "Sphere",
    "Torus"
    ]
#
# Length (for looping)
onll = len(onl)
#
#
# Object locations
# OLL
oll = [
    (-1.6256, 1.6256, 0.8128),
    (-3.8100, 0.0000, 0.0000),
    (-2.5400, 0.0000, 0.0000),
    (-1.2700, 0.0000, 0.0000),
    (1.2700, 0.0000, 0.0000),
    (2.5400, 0.0000, 0.0000),
    (3.8100, 0.0000, 0.0000)
    ]
#
#
# Object dimensions
# ODL
odl = [
    (1.4540, 1.3828, 1.6256),
    (0.6096, 0.6096, 0.6096),
    (0.6096, 0.6096, 0.6096),
    (0.0508, 0.0508, 1.2192),
    (0.6096, 0.6096, 0.0000),
    (0.6096, 0.6096, 0.6096),
    (0.7620, 0.7620, 0.1524)
    ]
#
#
# Object reference
# O
o = [
    0,
    0,
    0,
    0,
    0,
    0,
    0
    ]
#
#
# The material names listed here are those used in
# the "Color_ball" icosphere. Therefore they are
# available for usage in this script. I created
# this Object so that I could have common colors
# used in most of my projects.
#
# Materials
mats = [
    "Aqua",
    "Black",
    "Blue",
    "Brown",
    "Chocolate",
    "Clear",
    "Gray",
    "Green",
    "Lawn",
    "Magenta",
    "Mirror",
    "Orange",
    "Pink-hot",
    "Red",
    "Sky",
    "Tan",
    "Violet",
    "White",
    "Wood",
    "Yellow"
    ]
#
#
# In case we need it, I have also stored the length
# (number of items) in the list. This works well if
# using random numbers to get different materials.
#
# Len
matlen = len(mats) - 1
#
#
#
#
#
#
#
#
# Functions summary:
#
# The following functions are available within this
# script. Although they are not used much for the
# tutorial, so code can be explained step by step,
# they show how to build and use functions in
# Python. Using functions can greatly reduce the
# amount of code written as well as standardize
# how proceses are performed.
#
#
# Deselect All Objects
#dao()
#
#
# Select An Object
#   SAOO=Object SAOS=Select
#sao(saoo, saos)
#
#
# Make Object Active
#   MOAO=Object MOAS=Select
#moa(moao, moas)
#
#
# Move To Location
#   MTLO=Object MTLX=X MTLY=Y MTLZ=Z
#mtl(mtlo, mtlx, mtly, mtlz)
#
#
# Rotate To Angle
#   RTAO=Object RTAA=Axis RTAR=Radians RTAS=Set
#rta(rtao, rtaa, rtar, rtas)
#
#
# Scale To Size
#   STSO=Object STSX=X STSY=Y STSZ=Z STSA=Apply
#sts(stso, stsx, stsy, stsz, stsa)
#
#
# Set Object Dimensions
#   SODO=Object SODX=X SODY=Y SODZ=Z SODA=Apply
#sod(sodo, sodx, sody, sodz, soda)
#
#
# Create a collection
#   CACN=Name
#cref = cac(cacn)
#
#
# Move To Collection
#   MTCN=Name of Collection  MTCO=Object to move
#mtc(mtcn, mtco)
#
#
# Parent/Join
#   PJC=Child PJP=Parent PJW: 0(Parent) 1(Join) PJS: 0(No) 1(Yes)
#pj(pjc, pjp, pjw, pjs)
#
#
# Apply Material To Object
#   AMTOO = Object
#   AMTOM = Material name
#amto(amtoo, amtom)
#
#
# Update Display
#ud()
#
#
# Keyframe Location
#   KFLO=Object KFLF=Frame
#kfl(kflo, kflf)
#
#
# Keyframe Rotation
#   KFLO=Object KFLF=Frame
#kfr(kflo, kflf)
#
#
# Keyframe Scale
#   KFSO=Object KFSF=Frame
#kfs(kfso, kfsf)
#
#
# Get Random Number
#   RS=Start RE=End RP=Picked
#   Returns: # = RS to RE inclusive
#rn = grn(rs, re)
#
#
# Convert Radian to Degree
#     Radian = Degree * PI/180 (0.017453293)
#     Degree = Radian * 180/PI (57.2958)
#   CRTDV=Value CRTDC=Conversion
#crtd(crtdv, crtdc)
#
#
# Convert Inch to Meter (Blender)
#     Meter = Inch / 39.37
#     Inch = Meter * 39.37
#   CITMV=Value CITMC=Conversion
#citm(citmv, citmc)
#
#
#
#
#
#
#
#
# Functions:
#
#
# The following functions are provided to help
# understand how functions work in Python. These
# may be used if desired. They may also be
# modified to suit specific needs.
#
#
# Function
#
#
#
#
# Deselect All Objects
#dao()
#
# DAO
def dao():
    #
    # Deselect all
    bpy.ops.object.select_all(action='DESELECT')
    #
    #
    # End dao
#
#
#
#
# Select An Object
#   SAOO = Object
#   SAOS = Select: 0=Deselect 1=Select
#sao(saoo, saos)
#
# SAO
def sao(saoo, saos):
    #
    # Select?
    if saos == 1:
        #
        # Select
        saoo.select_set(True)
        #
        # Done
    else:
        #
        # Deselect
        saoo.select_set(False)
        #
        # Done
    #
    #
    # End sao
#
#
#
#
# Make Object Active
#   MOAO = Object
#   MOAS = Select: 0=Inactive 1=Active
#moa(moao, moas)
#
# MOA
def moa(moao, moas):
    #
    # Active?
    if moas == 1:
        #
        # Active
        bpy.context.view_layer.objects.active = moao
        #
        # Done
    else:
        #
        # Inactive
        bpy.context.view_layer.objects.active = None
        #
        # Done
    #
    #
    # End moa
#
#
#
#
# Move To Location
#   MTLO = Object
#   MTLX = X
#   MTLY = Y
#   MTLZ = Z
#mtl(mtlo, mtlx, mtly, mtlz)
#
# MTL
def mtl(mtlo, mtlx, mtly, mtlz):
    #
    # XYZ
    mtlo.location = (mtlx, mtly, mtlz)
    #
    # OR
    #
    # X
    #mtlo.location.x = mtlx
    #
    # Y
    #mtlo.location.y = mtly
    #
    # Z
    #mtlo.location.z = mtlz
    #
    # OR
    #
    # X
    #mtlo.location[0] = mtlx
    #
    # Y
    #mtlo.location[1] = mtly
    #
    # Z
    #mtlo.location[2] = mtlz
    #
    #
    # End mtl
#
#
#
#
# Rotate To Angle
#   RTAO = Object
#   RTAA = Axis
#   RTAR = Radians
#   RTAS = Set (Apply): 0=no 1=yes (assumes object is active)
#rta(rtao, rtaa, rtar, rtas)
#
# RTA
def rta(rtao, rtaa, rtar, rtas):
    #
    # Rotate
    rtao.rotation_euler[rtaa] = rtar
    #
    #
    # Apply?
    if rtas == 1:
        #
        # Apply
        bpy.ops.object.transform_apply(location=False, rotation=True, scale=False)
        #
        # Done
    #
    #
    # End rta
#
#
#
#
# Scale To Size
#   STSO = Object
#   STSX = X
#   STSY = Y
#   STSZ = Z
#   STSA = Apply: 0=no 1=yes (assumes object is active)
#sts(stso, stsx, stsy, stsz, stsa)
#
# STS
def sts(stso, stsx, stsy, stsz, stsa):
    #
    # XYZ
    stso.scale = (mtlx, mtly, mtlz)
    #
    # OR
    #
    # X
    #stso.scale.x = stsx
    #
    # Y
    #stso.scale.y = stsy
    #
    # Z
    #stso.scale.z = stsz
    #
    # OR
    #
    # X
    #stso.scale[0] = stsx
    #
    # Y
    #stso.scale[1] = stsy
    #
    # Z
    #stso.scale[2] = stsz
    #
    #
    # Apply?
    if stsa == 1:
        #
        # Apply
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
        #
        # Done
    #
    #
    # End sts
#
#
#
#
# Set Object Dimensions
#   SODO = Object
#   SODX = X
#   SODY = Y
#   SODZ = Z
#   SODA = Apply: 0=no 1=yes (assumes object is active)
#sod(sodo, sodx, sody, sodz, soda)
#
# SOD
def sod(sodo, sodx, sody, sodz, soda):
    #
    # XYZ
    sodo.dimensions = (mtlx, mtly, mtlz)
    #
    # OR
    #
    # X
    #sodo.dimensions.x = sodx
    #
    # Y
    #sodo.dimensions.y = sody
    #
    # Z
    #sodo.dimensions.z = sodz
    #
    # OR
    #
    # X
    #sodo.dimensions[0] = sodx
    #
    # Y
    #sodo.dimensions[1] = sody
    #
    # Z
    #sodo.dimensions[2] = sodz
    #
    #
    # Apply?
    if soda == 1:
        #
        # Apply
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
        #
        # Done
    #
    #
    # End sod
#
#
#
#
# Set Object Origin
#   SOOO = Object
#   SOOX = X
#   SOOY = Y
#   SOOZ = Z
#   SOOS = Select: 0=no 1=yes
#soo(sooo, soox, sooy, sooz, soos)
#
# SOO
def soo(sooo, soox, sooy, sooz, soos):
    #
    # Save current 3D Cursor location
    # Save
    c3dl = bpy.context.scene.cursor.location
    #
    # Set 3D cursor
    bpy.context.scene.cursor.location = (soox, sooy, sooz)
    #
    # Select?
    if soos == 1:
        #
        # Deselect All
        dao()
        #
        # Select
        sao(sooo, 1)
        #
        # Active
        moa(sooo, 1)
        #
        # Done
    #
    # Set origin
    bpy.ops.object.origin_set(type='ORIGIN_CURSOR')
    #
    # Reset 3D cursor
    bpy.context.scene.cursor.location = c3dl
    #
    # Select?
    if soos == 1:
        #
        # Deselect All
        dao()
        #
        # Active
        moa(sooo, 0)
        #
        # Done
    #
    #
    # End soo
#
#
#
#
# Create a collection
#   CACN = Name
#cref = cac(cacn)
#
# CAC
def cac(cacn):
    #
    # Exists?
    colref = bpy.data.collections.get(cacn)
    #
    # None
    if not colref:
        #
        # Create
        colref = bpy.data.collections.new(cacn)
        #
        # Link
        bpy.context.scene.collection.children.link(colref)
        #
        # Done
    #
    #
    # Return
    return colref
    #
    #
    # End cac
#
#
#
#
# Move To Collection
#   MTCN = Name of Collection
#   MTCO = Object to move
#mtc(mtcn, mtco)
#
# MTC
def mtc(mtcn, mtco):
    #
    # Collection
    colref = bpy.data.collections[mtcn]
    #
    # Not in?
    if mtco.name not in colref.objects:
        #
        # Unlink
        for colitsin in mtco.users_collection:
            #
            # Unlink
            colitsin.objects.unlink(mtco)
        #
        # Link
        colref.objects.link(mtco)
        #
        # Done
    #
    #
    # End mtc
#
#
#
#
# Parent/Join
#   PJC = Child
#   PJP = Parent
#   PJW = 0=Parent, 1=Join
#   PJS = 0=No, 1=Yes
#pj(pjc, pjp, pjw, pjs)
#
# PJ
def pj(pjc, pjp, pjw, pjs):
    #
    # Select?
    if pjs == 1:
        #
        # Deselect all
        bpy.ops.object.select_all(action='DESELECT')
        #
        # Child
        pjc.select_set(True)
        #
        # Parent
        pjp.select_set(True)
        #
        # Active
        bpy.context.view_layer.objects.active = pjp
        #
        # Done
    #
    #
    # What operation
    if pjw == 0:
        #
        # Parent
        bpy.ops.object.parent_set(type='OBJECT')
        #
        # Done
    else:
        #
        # Join
        bpy.ops.object.join()
        #
        # Done
    #
    #
    # Deselect?
    if pjs == 1:
        #
        # Child
        pjc.select_set(False)
        #
        # Parent
        pjp.select_set(False)
        #
        # Active
        bpy.context.view_layer.objects.active = None
        #
        # Done
    #
    #
    # End pj
#
#
#
#
# Apply Material To Object
#   AMTOO = Object
#   AMTOM = Material name
#amto(amtoo, amtom)
#
# AMTO
def amto(amtoo, amtom):
    #
    # Get material
    matref = bpy.data.materials.get(amtom)
    #
    # If exists
    if amtoo.data.materials:
        #
        # 1st slot
        amtoo.data.materials[0] = matref
        #
        # Done
    else:
        #
        # Create
        amtoo.data.materials.append(matref)
        #
        # Done
    #
    #
    # End amto
#
#
#
#
# Update Display
#ud()
#
# UD
def ud():
    #
    # Update
    bpy.context.view_layer.update()
    #
    #
    # End ud
#
#
#
#
# Keyframe Location
#   KFLO = Object
#   KFLF = Frame
#kfl(kflo, kflf)
#
# KFL
def kfl(kflo, kflf):
    #
    # Location
    kflo.keyframe_insert(data_path="location", frame = kflf)
    #
    #
    # End kfl
#
#
#
#
# Keyframe Rotation
#   KFRO = Object
#   KFRF = Frame
#kfr(kfro, kfrf)
#
# KFR
def kfr(kfro, kfrf):
    #
    # Rotation
    kfro.keyframe_insert(data_path='rotation_euler', frame = kfrf)
    #
    #
    # End kfr
#
#
#
#
# Keyframe Scale
#   KFSO = Object
#   KFSF = Frame
#kfs(kfso, kfsf)
#
# KFS
def kfs(kfso, kfsf):
    #
    # Scale
    kfso.keyframe_insert(data_path='scale', frame = kfsf)
    #
    #
    # End kfs
#
#
#
#
# Get Random Number
#   RS = Start
#   RE = End
#   RP = RS to RE inclusive
#rn = grn(0, 7)
#
# GRN
def grn(rs, re):
    #
    # Picked
    rp = random.randint(rs, re)
    #
    # Send
    return rp
    #
    #
    # End grn
#
#
#
#
# Convert Radian to Degree
#   CRTDV = Value
#   CRTDC = Conversion: 0=To_Radian 1=To_Degree
#crtd(crtdv, crtdc)
#
# CRTD
def crtd(crtdv, crtdc):
    #
    # Radian?
    if crtdc == 0:
        #
        # Radian = Degree * PI/180
        #        = Degree * 0.017453293
        #
        # To Radian
        return crtdv * 0.017453293
        #
        # Done
    else:
        #
        # Degree = Radian * 180/PI
        #        = Radian * 57.2957795
        #
        # To Degree
        return crtdv * 57.2957795
        #
        # Done
    #
    #
    # End crtd
#
#
#
#
# Convert Inch to Meter (Blender)
#   CITMV = Value
#   CITMC = Conversion: 0=To_Meter 1=To_Inch
#citm(citmv, citmc)
#
# CITM
def citm(citmv, citmc):
    #
    # Meter?
    if citmc == 0:
        #
        # Meter = Inch / 39.37
        #
        # To Meter
        return citmv / 39.37
        #
        # Done
    else:
        #
        # Inch = Meter * 39.37
        #
        # To Inch
        return citmv * 39.37
        #
        # Done
    #
    #
    # End citm
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
# Load Object reference into a list:
#
# Existing object names in the 'ONL' list are:
#   "Color_ball"  An Icosphere with many materials)
#   "Cone"        Loc: -150",0,0  Dim: 24",24",24"
#   "Cube"        Loc: -100",0,0  Dim: 24",24",24"
#   "Cylinder"    Loc: - 50",0,0  Dim: 24",24",24"
#   "Plane"       Loc:   50",0,0  Dim: 24",24", 0"
#   "Sphere"      Loc:  100",0,0  Dim: 24",24",24"
#   "Torus"       Loc:  150",0,0  Dim: 30",30", 6"
#
#
# NOTICE:
#     In case the script is being run again, we
#     need to reset the many original values.
#     This code does some of that.
#
# Rename "Ramp" back to "Plane"
if 'Ramp' in bpy.context.scene.objects:
    #
    # Get
    obj = bpy.data.objects["Ramp"]
    #
    # Rename
    obj.name = "Plane"
#
# Remove Objects from "Custom"
colref = bpy.data.collections['Custom']
#
# Each
for ob in colref.objects:
    bpy.data.objects.remove(ob, do_unlink = True)
#
# Remove Objects from "Empties"
colref = bpy.data.collections['Empties']
#
# Each
for ob in colref.objects:
    bpy.data.objects.remove(ob, do_unlink = True)
#
# Remove Objects from "Temp"
colref = bpy.data.collections['Temp']
#
# Each
for ob in colref.objects:
    bpy.data.objects.remove(ob, do_unlink = True)
#
# Remove Objects from "Pyramids"
colref = bpy.data.collections['Pyramids']
#
# Each
for ob in colref.objects:
    bpy.data.objects.remove(ob, do_unlink = True)
#
#
# We can get the Objects by name and store a reference
# to that Object. This allows us to get an Object using:
#     o[#]
#    Example:
#     o[3].location = (0, 0, 0)
#
# Objects
#   Process each entry in the "onl" list
for l in range(0, onll):
    #
    # Get Object by name
    obj = bpy.data.objects[onl[l]]
    #
    # Store its reference in 'o' list
    o[l] = obj
    #
    #
    # Report
    #print("Name:"+obj.name+" Loc:"+str(obj.location)+" Dim:"+str(obj.dimensions))
    #
    # In case the script is being run again, we
    # want to reset our original Object properties.
    #
    # Original location
    obj.location = oll[l]
    #
    # Original rotation
    obj.rotation_euler = (0, 0, 0)
    #
    # Original scale
    obj.scale = (1, 1, 1)
    #
    # Original dimension
    obj.dimensions = odl[l]
    #
    #
    # End l
#
#
# So now:
#   o[0] is the Icosphere object ("Color_ball")
#   o[1] is the Cone object
#   o[2] is the Cube object
#   o[3] is the Cylinder object
#   o[4] is the Plane object
#   o[5] is the Sphere object
#   o[6] is the Torus object
# There is no requirement to store Object references
#  in this manner. But it does provide an easy way
#  to get an Object, so I am using it for this
#  document to avoid using numerous variable names.
#
#
#
#
#
#
#
#
# Renaming:
#
# Lets rename the 'Plane' to 'Ramp'. We can perform
# many actions like this without having to "Select" an
# Object or make it "Active". Note that this only
# changes the Object name, but we will still be able
# to use the 'o[4]' reference to the same Object.
#
# Rename
o[4].name = "Ramp"
#
#
#
#
#
#
#
#
# Transforming:
#
# Values are entered in the units as follows:
#   Location specified in Meters
#     Convert Inch to Meters by:  Inch / 39.37
#   Rotation specified in Radians
#     Convert Degree to Radians by:  Degree * 0.017453293
#   Scale is specified in percentage
#     1.0=100% (same size)
#      .5=50% (half size)
#     2.0=200% (double size)
# This script includes converter functions
#
# NOTICE:
#   This document focuses on Euler rotation. If
#   preferred, Quaternion rotation can be used
#   instead by modification of related lines of
#   code.
#
# Here we have entered all three axis values in one
# statement. But we can access each axis on its own:
#    X axis using .x
#      o[3].location.x = -1.27
#    X axis using [0]
#      o[3].rotation_euler[0] = 0
#    X and Y axis using .xy
#      o[3].scale.xy = (.25, .25)
#    X, Y, and Z axis by default
#      o[3].scale = (.25, .25, 4)
#  Just remember that:
#    X axis is:   .x  or  [0]
#    Y axis is:   .y  or  [1]
#    Z axis is:   .z  or  [2]
#
#
# Location
#    X   Y    Z
#  -50", 0", 12"
o[4].location = (1.27, 0, 0.3048)
#
# Scale
#    X   Y   Z
#  .75, .75, 1
o[4].scale = (0.75, 0.75, 1)
#
# Rotation
#   X   Y   Z
#  0*, 45*, 0*
o[4].rotation_euler = (0, 0.785398163, 0)
#
#
#
#
#
#
#
#
# "Apply" the transformation:
#
# By applying transformation values, we are telling
# Blender that we want the current values to become
# the 'default' settings for the Object. This operation
# affects the internal 'transformation matrix' rather
# than the visual/physical appearance of the Object.
# It resets (clears) the transformation values.
#
# Applying the location causes the Object origin to be
# reset to (0,0,0). Since the Object origin is used for
# several operations (such as rotation pivot), this will
# affect how those operations interact with the Object
# after the apply.
#
# Applying the rotation clears the rotation transform
# values to (0,0,0) and recalculates the dimensions of
# the Object.
#
# Applying the scale resets the scale values to (1,1,1)
# and recalculates the dimensions according to the former
# scale values.
#
# Doing an APPLY after a transform is not always
# required. Applying before involking some operations
# like modifiers and constraints is usually needed in
# order to get appropriate results.
#
# Because it takes an 'operation' (notice the ".ops"
# in the line below) to perform an apply, we must
# first insure that the Object we want the operation
# to use is both 'selected' and 'active'. Operations
# normally work on the 'Active' Object rather than
# an Object reference specified on the code line.
# So we take care of that first.
#
# Deselect all Objects
bpy.ops.object.select_all(action='DESELECT')
#
# Select our 'Ramp'
o[4].select_set(True)
#
# Make the 'Ramp' Active
bpy.context.view_layer.objects.active = o[4]
#
# So now we can perform the 'Apply' operation. Here we
# are applying all transforms using values of "True".
# If we do not want to apply a particular transform,
# we simply set it to "False". So we are NOT going to
# apply the location.
#
# Apply all transforms
#bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)
#
# Because we just made our 'Ramp' Object 'selected'
# and 'active', we may often unset it after doing
# the operation. Doing so helps to prevent us from
# accidentally involving it in future operations.
# But since we will be setting other Objects in the
# next section, we will leave it as is here.
#
#
#
#
#
#
#
#
# Finding Objects:
#
# Warning:
#   Since Python can work with both inactive and
#   unselected Object references, there is no
#   guarantee that using bpy.context.object,
#   bpy.context.active_object or
#   bpy.context.selected_objects will point to or
#   include an Object which was just modified!
#
# Warning:
#   Even if all Objects are deselected in the Blender
#   GUI interface, there is still a reference to the
#   last (de)selected Object in bpy.context.object
#   and bpy.context.active_object! So if we need to
#   process an active Object, we should check to
#   make sure that it is 'Selected' before we do
#   actual processing!
#
# Although we can use Python code to work with Objects
# which are not selected nor active, sometimes we want
# to know which Object(s) are selected or active so that
# we can perform an operation on them specifically.
# This is useful for add-ons or custom operation code.
#
# We can get that information easily. The Blender API
# provides us with several ways to obtain reference
# to an Object(s) according to inquiries. This gives
# us a way to affect a change on one or many Objects
# at a time. For example, we can move the Y axis
# location of desired object(s) by 1 Meter in the
# examples below.
#
# For this section, we will 'select' the "Color_ball"
# Object and make it 'active'. This is our only
# pre-defined Object which is NOT in the "Items"
# collection, so this way it can still be involved
# in this sections activity.
#
# Deselect all Objects
bpy.ops.object.select_all(action='DESELECT')
#
# Select the "Color_ball"
o[0].select_set(True)
#
# Make the "Color_ball" Active
bpy.context.view_layer.objects.active = o[0]
#
#
# In the default .blend file, the 6 primitives were
# all lined up on the X axis with values of 0 for
# their Y axis locations. Now we will find (get a
# list of Objects that meet certain criteria)
# Objects and move them forward (negative Y axis
# values).
#
#
# The "current" Object will often be the same as
# the 'active' Object. In this case, it will be
# our "Color_ball" since we made that 'active'.
#
# Current Object:
#   bpy.context.object
ob = bpy.context.object
ob.location.y -= .25
#
#
# The "active" Object is (as it implies) the
# 'active' Object. In this case, it will be
# our "Color_ball" since we made that 'active'.
#
# Active Object:
#   bpy.context.active_object
ob = bpy.context.active_object
ob.location.y -= .25
#
#
# The "selected_objects" will be a list of zero or
# more Objects based upon how many Objects are
# 'selected' at the time. In this case, it will
# only be our "Color_ball" Object.
#
# All selected Objects:
#   bpy.context.selected_objects:
for ob in bpy.context.selected_objects:
    ob.location.y -= .5
#
#
# We can get the reference to any Object by
# looking it up by its name. In this case,
# we will find the "Cube" and move it forward.
#
# By Object name:
#   Change 'Cube' to desired Object name
ob = bpy.data.objects['Cube']
ob.location.y -= 1
#
#
# We can also get a list of all Objects located
# within a specific collection. In this case, all
# of our 'primitives' will be moved forward.
#
# In a certain collection:
#   Replace 'Items' with desired collection name
colref = bpy.data.collections['Items']
# No recursion
for ob in colref.objects:
    ob.location.y -= 1
# Includes children
#for ob in colref.all_objects:
#    ob.location.x += 1
#
#
# We can also get a list of ALL of the Objects in
# our project. Note that this include cameras,
# lights, and other Objects which may not be
# desired to affect. So here we will not run this
# code example.
#
# All Objects:
#   Commented out as some are not mesh Objects
#for ob in bpy.data.objects:
#    ob.location.x += 1
#
#
# However, we can limit our search to only include
# Objects with 'mesh' data. This would exclude the
# camera and light Objects.
#
# Only mesh Objects:
for ob in bpy.data.objects:
    if ob.type=='MESH':
        ob.location.y -= 1
#
#
# Now just to be safe, we will deselect all of
# the Objects and clear the active Object.
#
# Deselect all Objects
bpy.ops.object.select_all(action='DESELECT')
#
# Clear the Active Object
bpy.context.view_layer.objects.active = None
#
#
#
#
#
#
#
#
# Moving the 3D cursor:
#
# Since the Blender 3D Cursor is used for certain
# operations, there are times we would like to set
# its location. This is done just like setting an
# Object location, except we identify the Cursor
# as the target.
#
# Since this changes the 3D Cursor location, we can
# optionally save its location so that we can return
# it back after we are finished.
#
# Save current 3D Cursor location
c3dl = bpy.context.scene.cursor.location
#
# Ordinarily, the 3D Curson lives at the world
# origin. So we can easily reset it there after
# we have moved it elsewhere.
#
# Reset 3D cursor
bpy.context.scene.cursor.location = (0, 0, 0)
#
# Or we can restore the previously saved location.
#
# Restore 3D cursor former location
bpy.context.scene.cursor.location = c3dl
#
#
# We can set the values together or individually.
# So lets set it to a point we have calculated.
#
# Calculation results
#
# X
x = 1.1083566
#
# Y
y = -2.2286
#
# Z
z = -0.4664465
#
# Set 3D cursor
bpy.context.scene.cursor.location = (x, y, z)
#
#
#
#
#
#
#
#
# Changing Object Origins:
#
# The origin of an Object plays a vital role in many
# aspects of its existance. So it makes sense that
# we may need to change its origin sometimes. While
# there are several ways to do this, I prefer to use
# the 3D Cursor as the 'target'. We simply have to
# calculate the coordinate point where the origin
# will be, set the 3D Cursor to that point, and then
# invoke the operation.
#
# In the section above, we set the 3D Cursor location
# to a calculated point. That point is actually where
# the left and top corner of the 'Ramp' Object is at.
# Since we have already set the Cursor to that point,
# we can just use it to set the 'Ramp's origin. This
# allows 'local axis' rotation to occur at that point,
# instead of from the center of the Object.
#
# Insure the 'Ramp' is selected and active.
#
# Deselect all Objects
bpy.ops.object.select_all(action='DESELECT')
#
# Select the "Ramp"
o[4].select_set(True)
#
# Make the "Ramp" Active
bpy.context.view_layer.objects.active = o[4]
#
# Set the 'Ramp' origin to 3D Cursor location
#
# Origin set operation
bpy.ops.object.origin_set(type='ORIGIN_CURSOR')
#
# Restore the 3D Cursor
#
# Reset the 3D cursor to 0,0,0
bpy.context.scene.cursor.location = (0, 0, 0)
#
# Deselect everything and clear active
#
# Deselect all Objects
bpy.ops.object.select_all(action='DESELECT')
#
# Clear the Active Object
bpy.context.view_layer.objects.active = None
#
#
#
#
#
#
#
#
# Creating an Empty:
#
# Sometimes we want to reference a particular point
# of the coordinate system without using a physical
# Object to disrupt our scene. We could do this by
# using variables to store location, rotation, and
# scale values, but that does not provide a visible
# screen identity during our work. An 'Empty' is a
# reference Object which is visible during our work
# but does NOT affect rendered output. The empty
# acts much like a real Object as it has properties
# for location, rotation, and scale. So an empty is
# easier to manage than a set of custom variables
# and also gives us a visible indication of its
# location.
#
# We create an Empty much like any other Object.
#
# Create empty, store reference in emt
emt = bpy.data.objects.new("empty", None)
#
# Get "Empties" Collection
colref = bpy.data.collections['Empties']
#
# Link the empty to it
colref.objects.link(emt)
#
#
# Rename the empty
emt.name = "E.000"
#
#
# We can also set how it is displayed.
#
# Size
emt.empty_display_size = 2
#
# Type
emt.empty_display_type = 'PLAIN_AXES'
#
#
# We can set its location as usual. Here we have used
# a point just above where the 'Torus' Object was
# originally at. We will use this empty to modify the
# properties of the 'Torus'.
#
# Location
emt.location = (3.81, 0, 0.3048)
#
#
#
#
#
#
#
#
# Parenting Objects:
#
# Making a parent/child relationship between two
# (or more) Objects is a common task. It is also
# very easy to do in a script.
#
# The process works exactly like it does in the
# GUI interface. Select Object(s) which will be
# the children and then select the parent Object.
# In our script, we also have to set the parent
# Object as active.
#
# We will make the 'Torus' the child of the
# 'E.000' empty we created in the last section.
#
# First we deselect all Objects.
bpy.ops.object.select_all(action='DESELECT')
#
# Then we select both Objects.
#
# Child
o[6].select_set(True)
#
# Parent
emt.select_set(True)
#
# And then make the empty active.
#
# Active
bpy.context.view_layer.objects.active = emt
#
# Now we can run the operation. The 'type' of
# relationship can be changed if desired.
#
# Parent
bpy.ops.object.parent_set(type='OBJECT')
#
#
# After the operation I usually deselect at
# least the child Object(s), because we would
# be more likely to continue work on the parent.
# But we could also deselect both of them and
# clear the active Object if we desire to.
#
# Deselect child
o[6].select_set(False)
#
# Deselect parent
emt.select_set(False)
#
# Clear active
bpy.context.view_layer.objects.active = None
#
#
#
#
#
#
#
#
# Joining Objects:
#
# Joining two (or more) Objects makes them be
# physically attached as a single Object.
# The selected Object(s) are treated as the
# children and the active Object is the
# Parent. The resulting single Object will
# have the parent Object name and origin.
#
# Like parenting, any location or distance
# of Object(s) is allowed in joining. But
# with parenting, a child can be transformed
# separate from the parent. When we 'join'
# Objects, we can no longer reposition the
# children independently. Thus it is very
# important to have all involved Objects
# to be in place prior to joining them! So
# we will set the location of our Objects
# before we run the operation.
#
# For this section, we are going to create
# a cube and a cone using a normal operation
# (ops) method. We will place the cone on top
# of the cube and then join them using the
# cube as the 'parent'.
#
# We will use the "Temp" collection
#
# Collection
colref = bpy.data.collections["Temp"]
#
#
# Add a Cone
#bpy.ops.mesh.primitive_cone_add(vertices=32, radius1=1.0, radius2=0.0, depth=2.0, end_fill_type='NGON', calc_uvs=True, enter_editmode=False, align='WORLD', location=(0.0, 0.0, 0.0), rotation=(0.0, 0.0, 0.0), scale=(0.0, 0.0, 0.0))
bpy.ops.mesh.primitive_cone_add()
#
# Store reference
jcone = bpy.context.object
#
# Link to 'Temp' collection
#colref.objects.link(jcone)
#
# Name
jcone.name = "J_Cone"
#
# Location
jcone.location = (2.0, 2.0, 0.6)
#
# Scale
jcone.scale = (0.2, 0.2, 0.2)
#
#
# Add a Cube
#bpy.ops.mesh.primitive_cube_add(size=2.0, calc_uvs=True, enter_editmode=False, align='WORLD', location=(0.0, 0.0, 0.0), rotation=(0.0, 0.0, 0.0), scale=(0.0, 0.0, 0.0))
bpy.ops.mesh.primitive_cube_add()
#
# Store reference
jcube = bpy.context.object
#
# Link to 'Temp' collection
#colref.objects.link(jcube)
#
# Name
jcube.name = "J_Cube"
#
# Location
jcube.location = (2.0, 2.0, 0.2)
#
# Scale
jcube.scale = (0.2, 0.2, 0.2)
#
#
# Handle the selection of Objects
#
# Deselect all Objects
bpy.ops.object.select_all(action='DESELECT')
#
# Child
jcone.select_set(True)
#
# Parent
jcube.select_set(True)
#
# Active
bpy.context.view_layer.objects.active = jcube
#
#
# Join them
bpy.ops.object.join()
#
#
# The 'cone' object is deleted by the join operation.
# So there is no need to deselect it. Optionally, we
# can deselect the 'cube' and clear the active Object.
#
# Deselect
#jcube.select_set(False)
#
# Active
#bpy.context.view_layer.objects.active = None
#
#
#
#
#
#
#
#
# Use a Modifier:
#
# Scripts can do so much more than just creating and
# transforming Objects. We can also enable modifiers
# and constraints. In this example, we will construct
# a "Simple Deform" modifier for our "Cylinder" Object.
# We will set it to 'bend' the cylinder 90 degrees on
# the Y axis.
#
# The "Cylinder" Object in the included .blend file
# has already been prepared for this process with
# many 'loop cuts' to give it enough geometry for
# bending.
#
# Each modifier has its own options and default
# values. Using the Python tooltips or the Python
# Console (both as suggested at the beginning of
# this document) we can get the name of an option
# and often clues to what setting(s) are available.
#
# The creation of a "Simple Deform" modifier is a
# very long line of code. It includes a name so it
# can be referenced later on. It also specifies the
# type of modifier we want.
#bpy.data.objects["Cylinder"].modifiers.new(name="Bend_me", type="SIMPLE_DEFORM")
#
# But I like to store references in variables for
# easier access. Since we already have a reference
# for the "Cylinder" Object ('o[3]'), we can do this
# shortcut version instead.
#
# Create the modifier
sdm = o[3].modifiers.new(
    name='Bend_me',
    type='SIMPLE_DEFORM'
    )
#
# This creates a variable named 'SDM' (Simple Deform
# Modifier, use any variable name desired) which now
# holds the reference to the modifier. So now we do
# not need to remember the modifier name as we can
# just use 'sdm' to set any options. So this:
#
# Method
#bpy.data.objects['Cylinder'].modifiers["Bend_me"].deform_method = "BEND"
#
# becomes a more attractive:
#
# Method ("BEND" used, could be "TWIST", "TAPER", or "STRETCH")
sdm.deform_method = "BEND"
#
# We can set other options in a similar fashion.
#
# Angle (90*)
sdm.angle = 1.570796
#
# Axis (Y)
sdm.deform_axis = "Y"
#
#
# Originally I felt it would be very complex
# when I first considered adding a modifier in
# a script. But this one took only 4 short lines
# of code!
#
# If we want to 'apply' (make results permanent)
# the modifier, we must run an 'operation'. So
# we would need to insure our Object is the active
# one before running the next line.
#
# Apply
#bpy.ops.object.modifier_apply(modifier = sdm.name)
#
#
#
#
#
#
#
#
# Use a Constraint:
#
# In an earlier section, we created a 'Cube' and a
# 'Cone' which we then joined together. "Join"ing
# is a permanent action, so we cannot access those
# structures individually now. But we can obtain a
# similar result using a "Copy Location" contraint
# while keeping the ability to modify each Object
# on its own.
#
# Note that this example is only to copy the
# location of a target Object. If we wanted our
# Object to also respond to other transformations,
# we would also use a "Copy Rotation" and "Copy
# Scale" contraint, or use the all-in-one
# "Transforms" constraint.
#
# The "Cube" and "Cone" Objects included in the
# .blend file will be used here. We will set the
# constraint on the "Cone" so that it will follow
# the location of the 'target' ("Cube"). Moving
# the Cube will then also move the Cone.
#
# Because both the Cube and Cone have their origin
# at their center, if we copy the location now it
# will place the Cone inside of the Cube. So first
# we need to adjust the origin of the Cone so that
# it will sit on top of the Cube. Since both Objects
# are 24" high, we will LOWER the origin of the
# Cone by 24" (0.6096 meters).
#
# Get current location of Cone (o[1])
#
# X
x = o[1].location.x
#
# Y
y = o[1].location.y
#
# Z
z = o[1].location.z
#
#
# Adjust Z downward by 24", the distance of
# origin-to-edge (12") of both the "Cone" and
# the "Cube". This will make the "Cone" sit
# on top of the "Cube" once we change the
# "Cone" origin to this position.
#
# Z
z -= 0.6096
#
#
# Set the location of the 3D cursor
bpy.context.scene.cursor.location = (x, y, z)
#
#
# Reset the "Cone" Origin to the 3D Cursor
#
# Deselect all Objects
bpy.ops.object.select_all(action='DESELECT')
#
# Select the "Cone"
o[1].select_set(True)
#
# Make the "Cone" Active
bpy.context.view_layer.objects.active = o[1]
#
# Set the 'Cone' origin to 3D Cursor location
bpy.ops.object.origin_set(type='ORIGIN_CURSOR')
#
# Reset the 3D cursor to 0,0,0
bpy.context.scene.cursor.location = (0, 0, 0)
#
# We will keep the "Cone" selected and active.
#
#
# Now we can create a "Copy location" constraint on
# the "Cone" Object. As we did in the section above,
# we will use a shortcut path and store reference
# to the constraint with a variable named "CLC"
# (Copy Location Constraint).
#
# Create constraint
clc = o[1].constraints.new("COPY_LOCATION")
#
# Set the "Cube" as the 'target'
clc.target = o[2]
#
#
# Now moving the Cube will also move the Cone,
# which is sitting on top of it. If we try to
# move the Cone to a different location, it will
# not move while the contraint is still active.
#
#
# The advantage of using a constraint(s) to keep
# the Cone on top of the Cube is the flexibility.
# The "Join"ed version cannot be adjusted. But we
# can set the 'influence' value of the constraint
# to control how much it affects the action. The
# default value of 1.0 means the two Objects will
# have the same location values. A value of 0.0
# means there will be no adjustment to the Cone's
# original location. Any value between 0.0 and
# 1.0 will provide for varied levels of influence.
#
# Set the amount of influence
clc.influence = 1.0
#
#
#
#
#
#
#
#
# Mess with the Mesh:
#
# The "Sphere" told me that it had not been used
# in this tutorial. So I decided to 'lift' up its
# spirits with an example of how to mess around
# with the actual mesh of an Object.
#
# I first learned how to modify vertex data while
# trying to use "Hooks" (attaching a vertex to
# the location of an Empty). I learned how to
# modify vertex data 'directly' (through the
# 'data' property of an Object). Later on I
# discovered the same can be done with "Bmesh",
# Blenders internal mesh editing API. Honestly I
# do not know if Bmesh is safer or is preferred
# over direct access. Both methods seem to work
# well, in other words direct access is NOT
# prevented. For full disclosure, the example
# for Bmesh below is a slightly modified code
# from the Bmesh section of the Blender manual.
# The direct method example is a modified version
# from my "Dancing Cube" project where I used
# hooks to move the vertices of a Cube.
#
# Here we only cover vertex modifications. Edge
# and face data can also be manipulated, but if
# so there are recalculations which need to be
# managed. Since changing a vertex also changes
# edges and faces, I have been just letting
# Blender 'do the math' after I change a vertex
# and not messing with edge or face data.
#
# As will be seen in the examples below, both
# methods have very similar structure. Both
# require that we get a reference to the Object
# .data, which I store in the "SMD" variable.
# The only notable difference is that with Bmesh
# we are working with a copy of the data instead
# of working with it directly.
#
# Both methods loop through all of the vertices
# of the mesh data. I learned that we can get
# the following information:
#   v.index  Is the index #, 0 to (number_of_vertices)-1
#   v.co     Is the full coordinate:  Vector (0.0000, 0.0595, 0.2989)
#   v.co.x   Is the X coordinate:  0.0
#   v.co.y   Is the Y coordinate:  0.0595
#   v.co.z   Is the Z coordinate:  0.2989
# And we modify it in the usual way:
#   v.co.z = 0.257
#   v.co.z += 0.257
#
# Since there are two methods, I decided to use
# each to modify half of the "Sphere" Object.
# Using direct access, the vertices in the top
# half of the sphere are moved upward. Using
# Bmesh, the vertices in the bottom half of the
# sphere are moved to the right.
#
#
#
# Directly modify vertex data
#
# Get the "Sphere"s mesh data
smd = o[5].data
#
# Each vertex
for v in smd.vertices:
    #
    # If Z is above 0
    if v.co.z > 0:
        #
        # Raise it another .5 meters
        v.co.z += .5
        #
        # Done
    #
    # End V
#
#
#
# Modify data using Bmesh
#
# Get the "Sphere"s mesh data
smd = o[5].data
#
# Get a BMesh representation
bm = bmesh.new()
#
# fill it in from a Mesh
bm.from_mesh(smd)
#
# Modify the BMesh, can do anything here...
for v in bm.verts:
    #
    # If Z is below 0
    if v.co.z < 0:
        #
        # Move the vertex .2 meters to the right
        v.co.x += .2
        #
        # Done
    #
    # End V
#
# Finish up, write the bmesh back to the mesh
bm.to_mesh(smd)
#
# free and prevent further access
bm.free()
#
#
# Since I do not use mesh data manipulation very
# much, I can only provide the above information.
# However, later we will create some object using
# raw mesh data which may also be helpful.
#
#
#
#
#
#
#
#
# Lets text:
#
# While learning how to work with Python in Blender,
# I decided that I wanted my debugging text in the
# scene instead of just "print"ed to the console.
# So I wrote a little function to create a text
# Object and fill it with the text from a "print"
# statement. I find this very useful sometimes.
#
# So I decided to include this section about how
# to create a Text Object. Here I use it to show
# the name of the "Sphere" Object, but any text
# (or any type converted to a string) could be
# entered on the "tfc.body = " line.
#
# This is just the basic code to do the job. My
# function is much more complex as I can pass it
# a reference to an object for the text to be
# placed next to. For this tutorial, all we need
# is the basics.
#
#
# Reference to the "Temp" collection
colref = bpy.data.collections['Temp']
#
#
# Create a curve object for text
tfc = bpy.data.curves.new(type="FONT", name="Text Area")
#
# The actual text to show, anything we desire but must be a 'string'
tfc.body = o[5].name
#
#
# Create the text object with the Curve data
tfo = bpy.data.objects.new(name="Text Object", object_data=tfc)
#
#
# Link to "Temp" Collection
#bpy.context.scene.collection.objects.link(tfo)
colref.objects.link(tfo)
#
#
# Optionally convert to mesh
#
# Deselect all Objects
bpy.ops.object.select_all(action='DESELECT')
#
# Select
tfo.select_set(True)
#
# Active
bpy.context.view_layer.objects.active = tfo
#
# To mesh
bpy.ops.object.convert(target='MESH')
#
#
# Move to location
tfo.location = (2.54, -1.0, 0.0)
#
#
#
#
#
#
#
#
# Create a mesh Object from raw data:
#
#
# Define
#
# This creates a single mesh Object in a T shape.
#  Its structure is similar to four cubes.
#  It has 20 Vertices, 36 Edges, and 21 Faces
#
# The 'Verts' array has the actual coordinates of each Vertex.
#  It is a 3-value Tuple with the X, Y, and Z values.
#
# The 'Edges' array is a series of 2 value Tuples. Each of
#  these has the index value of the 2 Vertices which define
#  the two ends of an edge.
#
# The 'Faces' array is a series of 3 (or more) value Tuples.
#  Like the Edges, each of these has the index value of the
#  Vertices which define the edges of a face.
#
# We must be VERY CAREFUL when calculating and entering this
#  data! Even if only one Vertex value is wrong, it can have
#  a ripple effect on many surrounding Edges and Faces!
#
#
# Vertices
verts = [
    # Front
    #  Outer
    (-0.0508, -0.0508, 0),
    (-0.0508, -0.0508, 0.5588),
    (-0.6096, -0.0508, 0.5588),
    (-0.6096, -0.0508, 0.6096),
    (-0.0508, -0.0508, 0.6096),
    (0.0508, -0.0508, 0.6096),
    (0.6096, -0.0508, 0.6096),
    (0.6096, -0.0508, 0.5588),
    (0.0508, -0.0508, 0.5588),
    (0.0508, -0.0508, 0),
    # Back
    #  Outer
    (-0.0508, 0.0508, 0),
    (-0.0508, 0.0508, 0.5588),
    (-0.6096, 0.0508, 0.5588),
    (-0.6096, 0.0508, 0.6096),
    (-0.0508, 0.0508, 0.6096),
    (0.0508, 0.0508, 0.6096),
    (0.6096, 0.0508, 0.6096),
    (0.6096, 0.0508, 0.5588),
    (0.0508, 0.0508, 0.5588),
    (0.0508, 0.0508, 0)
]
#
# Edge
edges = [
    # Front
    #  Outer
    (0, 1),
    (1, 2),
    (2, 3),
    (3, 4),
    (4, 5),
    (5, 6),
    (6, 7),
    (7, 8),
    (8, 9),
    (9, 0),
    #  Inner
    (1, 4),
    (1, 8),
    (5, 8),
    # Sides
    (0, 10),
    (1, 11),
    (2, 12),
    (3, 13),
    (4, 14),
    (5, 15),
    (6, 16),
    (7, 17),
    (8, 18),
    (9, 19),
    # Back
    #  Outer
    (10, 11),
    (11, 12),
    (12, 13),
    (13, 14),
    (14, 15),
    (15, 16),
    (16, 17),
    (17, 18),
    (18, 19),
    (19, 10),
    #  Inner
    (11, 14),
    (11, 18),
    (15, 18)
]
#
# Face
faces = [
    # Top
    (0, 1, 8, 9),
    (1, 2, 3, 4),
    (1, 4, 5, 8),
    (5, 6, 7, 8),
    # Side
    #  Outer
    (0, 1, 11, 10),
    (1, 2, 12, 11),
    (2, 3, 13, 12),
    (3, 4, 14, 13),
    (4, 5, 15, 14),
    (5, 6, 16, 15),
    (6, 7, 17, 16),
    (7, 8, 18, 17),
    (8, 9, 19, 18),
    (9, 0, 10, 19),
    # Inner
    (1, 8, 18, 11),
    (1, 4, 14, 11),
    (5, 8, 18, 15),
    # Bottom
    (10, 11, 18, 19),
    (11, 12, 13, 14),
    (11, 14, 15, 18),
    (15, 16, 17, 18)
]
#
#
# Create the Mesh
#
# Add a new mesh
#   Change 'T_Mesh' to any name desired
mesh = bpy.data.meshes.new('T_Mesh')
#
# Install the data
mesh.from_pydata(verts, edges, faces)
#
# Update
mesh.update()
#
#
# Create the Object
#
# Add a new Object
#   Change 'My_T' to any name desired
customT = bpy.data.objects.new('My_T', mesh)
#
# Link
#   This would link (add) the Object in a non
#   specific place in the collection tree.
#bpy.context.collection.objects.link(object)
#
#   But we can grab a reference to an existing
#   collection and link the Object in there instead.
#
# Collection
colref = bpy.data.collections['Custom']
#
# Link
colref.objects.link(customT)
#
#
# We can also add a material to the Object. Just
# for fun, lets use 'Red', which is one of the
# colors in the 'Color_ball' Object.
#
# Get material
matref = bpy.data.materials.get('Red')
#
# Since we just created this Object from scratch,
# we know it has no materials yet. So we can just
# add one in.
#
# Append (add) to the Object.
customT.data.materials.append(matref)
#
#
# Move it forward, out of the way for future items.
customT.location = (0, -1, 0)
#
#
#
#
#
#
#
#
# Make a pyramid of Pyramids:
#
# It would be very possible to construct interesting
# mesh objects by using algorithms to calculate the
# vertices, edges, and faces. But since we can already
# easily create complex structures with the tools in
# Blender, we need to decide whether it is worthwhile
# to create the object from a lot of raw data or using
# GUI tools instead.
#
# However, we can use a combination of creating a
# 'primitive' (a simple geometric shape used as a
# basis to create a more complex Object) and then
# manipulate it via script. This makes sense as we
# are only using a small amount of raw data to
# define our mesh structure.
#
# We will create a very simple pyramid mesh
# primitive. Then we will repeatedly copy the mesh
# data into new Objects. As we go, we will set the
# Object locations to form a large pyramid with a
# very simple formula. Because both the mesh data
# and the positioning formula are very simple, it
# makes sense to handle it via script. Although
# this example could be created fairly easily using
# the GUI interface, we could modify the parameters
# of the script to make an extremely huge pyramid
# quickly which would take much longer to create
# using the GUI.
#
# Our simple mesh pyramid will be .2 x.2 x.2 Meters
# (about 8"x8"x8"). It consists of only 5 vertices,
# 8 edges, and 5 faces.
#
#
# First we will define some values.
#
# The 'PP' (Pyramid Positions) list holds all of the
# coordinate positions for our Objects. Since our
# structure will be symmetrical with the center on the
# 0,0,0 coordinate, these values work for both row
# and column placement. These values are in Meters.
#
# Pyramid Positions
pp = [
    -0.6,
    -0.5,
    -0.4,
    -0.3,
    -0.2,
    -0.1,
    0.0,
    0.1,
    0.2,
    0.3,
    0.4,
    0.5,
    0.6
]
#
#
# Next we can pre-define the names we will use for
# the mesh data ('Pyra_mesh'), the base Object names
# ('Pyra.') and get a reference to the collection
# where we will store everything we create.
#
# Names
#
# Mesh
pnm = 'Pyra_mesh'
#
# Object base
#   Will produce 'Pyra.000', 'Pyra.001', 'Pyra.002', ... 'Pyra.230'
pno = 'Pyra.'
#
# Collection
pcol = bpy.data.collections['Pyramids']
#
#
# Before we create anything, lets setup a material
# for our Objects. In the example above, we simply
# used an existing material. Sometimes we want to
# create new materials, often using 'shader nodes'.
# In this example, we will setup a random color
# material that provides different colors for each
# Object. That fits this example because we are
# going to create 231 pyramid Objects.
#
# Working with nodes is actually pretty easy. But
# with so many different combinations available it
# would be hard to cover them all. So here we are
# going to show how to use a specific node setup,
# but it can be adapted for use with other nodes.
#
# Our 'Random material' will be created using:
#   An 'Object Info' node
#   A 'Color Ramp' node
#   A 'Principled BSDF' node
#   And a 'Material Output' node
# If not familiar with using shader nodes, it
# would be a good idea to do some research on them.
#
#
# First we create a material and get its reference
rmat = bpy.data.materials.new(name = 'Rand')
#
# Enable it to 'Use nodes'
rmat.use_nodes = True
#
# Setup a shortcut for the node path
nodes = rmat.node_tree.nodes
#
# Get a reference to the 'Material Output' node
sno = nodes.get("Material Output")
#
# Get a reference to the 'Object Info' node
snoi = nodes.new(type='ShaderNodeObjectInfo')
#
# Get a reference to the 'Color Ramp' node
sncr = nodes.new(type='ShaderNodeValToRGB')
#
# This is how to adjust options for the Color Ramp.
# Here I have setup a full range spectrum with
# only the default first and last marker. These can
# be defined any way that is desired.
#
# Mode
sncr.color_ramp.color_mode = "HSL"
#
# Interpolation (CW=Clockwise)
sncr.color_ramp.hue_interpolation = "CW"
#
# Position of the first marker
sncr.color_ramp.elements[0].position = 0.000
#
# Color of the first marker
sncr.color_ramp.elements[0].color = (0.003, 0.816, 0.953, 1.000)
#
# Position of the last marker
sncr.color_ramp.elements[1].position = 1.000
#
# Color of the last marker
sncr.color_ramp.elements[1].color = (0.002, 1.000, 0.816, 1.000)
#
#
# Get a reference to the 'Principled BSDF' node
snpb = nodes.new(type='ShaderNodeBsdfPrincipled')
#
#
# Now we can link the nodes together.
#
# Setup a shortcut for the link path
links = rmat.node_tree.links
#
# Object Info to Color Ramp
snoiLink = links.new(snoi.outputs[5], sncr.inputs[0])
#
# Color Ramp 'Color' to Principled BSDF
sncrLink = links.new(sncr.outputs[0], snpb.inputs[0])
#
# We need to do this if we want the Alpha channel
# of the Color Ramp to have an impact on the scene!
#
# Color Ramp 'Alpha' to Principled BSDF
sncrLink = links.new(sncr.outputs[1], snpb.inputs[4])
#
# Principled BSDF to Material Output
snpbLink = links.new(snpb.outputs[0], sno.inputs[0])
#
#
# Define
#
# Now we can setup the data for our mesh. The mesh
# will be .2 x.2 x.2 (about 8"x8"x8"). We create it
# centered at the 0,0,0 coordinate so that our
# created Object(s) will have their origin at their
# bottom center. This will be useful later on.
#
# We define vertices, edges, and faces just like we
# did in the previous section.
#
# Vertices
pv = [
    (-0.1, -0.1, 0.0),
    (-0.1,  0.1, 0.0),
    ( 0.1,  0.1, 0.0),
    ( 0.1, -0.1, 0.0),
    ( 0.0,  0.0, 0.2)
]
#
#
# Edges
pe = [
    (0, 1),
    (1, 2),
    (2, 3),
    (3, 0),
    (0, 4),
    (1, 4),
    (2, 4),
    (3, 4)
]
#
#
# Faces
pf = [
    (0, 1, 2, 3),
    (0, 1, 4),
    (1, 2, 4),
    (2, 3, 4),
    (3, 0, 4)
]
#
#
# Now we can create the mesh with the data.
#
# Create a mesh
pmesh = bpy.data.meshes.new(pnm)
#
# Install the data
pmesh.from_pydata(pv, pe, pf)
#
# Update
pmesh.update()
#
#
# Next we will setup the limits of the loops (levels,
# rows, and columns) which will guide the placement
# of Objects during the construction of the model.
#
# In this example, we will have 7 levels, 7 rows, and
# 7 columns of pyramid primitives, arranged to make
# a large pyramid. We can change these limits for
# smaller or larger models. Because we set the limits
# here, we do not have to modify the building code
# below it when we change these limits.
#
#
# Levels, 0-6
#
# Current
pl = 0
#
# Max + 1
plm = 7
#
#
# Each row and column equals 1/2 the size of each
# Object. This is how we are able to offset the
# Objects for the odd numbered levels.
#
# Rows, 0-12
#
# Current
pr = 0
#
# Start
prs = 0
#
# End, 1 above max
pre = 13
#
# Increment (2 columns)
pri = 2
#
#
# Columns, 0-12
#
# Current
pc = 0
#
# Start
pcs = 0
#
# End, 1 above max
pce = 13
#
# Increment (2 columns)
pci = 2
#
#
# We also need to keep track of how many Objects
# have been created, so we can set the numeric value
# of the names (.000, .001, etc).
#
# Object counter
poc = 0
#
#
# Temporary location calculations
#
# X
x = 0
#
# Y
y = 0
#
# Z
z = 0
#
#
# For each level, we lift Objects upwards.
#
# Add Z
za = .2
#
#
# The rotation needed to flip Objects upside-down.
#
# 180*
r180 = 3.14159274
#
#
# Now we have the mesh data and the limits of the
# construction loops. So we can start creating and
# building! The 'for' loops manage the positions,
# letting us generate location data. So all we need
# to do is create Objects, link the mesh data, set
# its name, set its location, and add the material.
#
# Copy mesh to new Objects
#
# Levels
# 0 to plm
for pl in range(0,plm):
    #
    # Rows
    # prs to pre
    for pr in range(prs,pre,pri):
        #
        # Columns
        # pcs to pce
        for pc in range(pcs,pce,pci):
            #
            #
            # Add an Object
            #
            # Name
            #   Name base + number
            onam = pno + f"{poc:03}"
            #
            # New Object
            obj = bpy.data.objects.new(onam, pmesh)
            #
            # Link to collection
            pcol.objects.link(obj)
            #
            #
            # Calculate the location
            #
            # X
            x = pp[pr]
            #
            # Y
            y = pp[pc]
            #
            #
            # Move it!
            obj.location = (x, y, z)
            #
            #
            # Add the 'rand' material
            obj.data.materials.append(rmat)
            #
            #
            # Object counter
            poc += 1
            #
            #
            # After the first (bottom) level, we can
            # use the same mesh flipped upside-down
            # to fill in the gaps on the level below.
            # To see why we do this, comment out this
            # section and run the script again.
            #
            # Down?
            if pl > 0:
                #
                #
                # Add an Object
                #
                # Name
                onam = pno + f"{poc:03}"
                #
                # Object
                obj = bpy.data.objects.new(onam, pmesh)
                #
                # Link
                pcol.objects.link(obj)
                #
                #
                # Location
                #
                # X
                x = pp[pr]
                #
                # Y
                y = pp[pc]
                #
                #
                # Move
                obj.location = (x, y, z)
                #
                #
                # Rotate 180* to flip upside-down
                obj.rotation_euler[0] = r180
                #
                #
                # Add 'rand' material
                obj.data.materials.append(rmat)
                #
                #
                # Object counter
                poc += 1
                #
                #
                # End down
            #
            #
            # End pc
        #
        #
        # End pr
    #
    #
    # After each level is finished, we need to reset
    # the start/end positions for rows/columns. We
    # simply move 1 row/col inward, which staggers the
    # 2-row/column wide Objects accordingly.
    #
    # Adjust
    #
    # Row
    #
    # Start + 1
    prs += 1
    #
    # End - 1
    pre -= 1
    #
    # Col
    #
    # Start + 1
    pcs += 1
    #
    # End - 1
    pce -= 1
    #
    #
    # We also need to raise the Object the height of
    # the level below.
    #
    # Z
    z += za
    #
    #
    # End pl
#
#
# Finally, we update the context to insure that all of
# the work we just did is shown!
#
# Update the view layer
bpy.context.view_layer.update()
#
#
# Options:
#
# If desired, we can either 'parent' or 'join' these
# Objects together so they will act as a single model.
# This is easy to do by simply selecting all of the
# Objects in the 'Pyramids' collection and then making
# one of them active prior to the operation.
#
# But since we are going to have a bit of fun with
# this pyramid later on, we will leave it as is.
#
#
#
#
#
#
#
#
# Keyframes and animation:
#
# Now that we have created so many items in our
# scene, it is time to wrap it all into a great
# animation! Well, at least a basic animation.
# First, lets discuss some settings that might
# need to be adjusted. These are found in the
# "Scene" tab of the "Properties" editor. Set
# these as desired.
#
#
# Resolution
#
# Width (1920px)
#bpy.data.scenes["Scene"].render.resolution_x = 1920
#
# Height (1080px)
#bpy.data.scenes["Scene"].render.resolution_x = 1080
#
#
# Frame
#
# Current frame
#bpy.context.scene.frame_current = 0
#
# Start frame (1)
#bpy.data.scenes["Scene"].frame_start = 1
#
# End frame (600)
#bpy.data.scenes["Scene"].frame_end = 600
#
# Frame step (1)
#bpy.data.scenes["Scene"].frame_step = 1
#
# FPS (Frames Per Second) (24)
#
#
# Polation
#
# Extrapolation
#  Before first keyframe and after last keyframe
#   ('CONSTANT', 'LINEAR', 'BEZIER')
#bpy.ops.graph.extrapolation_type(type='LINEAR')
#
# Interpolation
#  Between keyframes
#   ('CONSTANT', 'LINEAR', 'BEZIER')
#bpy.ops.graph.interpolation_type(type='LINEAR')
#
#
# Easing/Smoothing
#
# Ease type
#   ('AUTO', 'EASE_IN', 'EASE_OUT', 'EASE_IN_OUT')
#bpy.ops.graph.easing_type(type='AUTO')
#
# Auto Smoothing
# If using 'Linear' interpolation, insure that
# the setting in:
#    Edit: Preferences: Animation: Default_Smoothing_Mode
# is set to "None". If set to "Continous Acceleration"
# the movement of animations may not be truly linear.
#
#
# Consider which extrapolation, interpolation
# and/or easing methods are the right ones for
# each project!
#
# "Constant" interpolation means that the full
# effect of the action takes place immediately
# and then stays there. "Linear" interpolation
# means that the effect is applied smoothly
# from the former keyframe. Our animation will
# use linear interpolation.
#
# "Easing" refers to a rate of change at the
# beginning or end of the effect. "Ease in" will
# start the effect slowly and gradually increase
# it. This would be useful for a car starting
# movement from being stopped. "Ease Out" will
# slow the effect near the end of the action.
# This would be useful for a car coming to a
# stop sign. "Ease In/Out" applies both effects
# just above. Our animation will not use easing,
# as we do NOT want the pyramid to slow down or
# speed up during the segments of its journey.
# We can mimic increasing speed as it travels
# down the side of the big pyramid by setting
# keyframes of differnt distances traveled or
# at different frame intervals.
#
#
#
# We can calculate how long our animation will be:
#  #_of_frames = End_frame - Start_frame
#  Time_in_seconds = #_of_frames / FPS
# So 600 frames at 24 FPS, we have a 25 second
# animation time. Often we will know how long we
# want an animation to be, so we can calculate
# how many frames it will need:
#   #_of_frames = #_of_seconds * Frames_per_second
#
#
# Advance planning of our animation usually helps
# to get things right without having to go back
# to adjust them several times. We should consider
# how many different actions occur and how long
# each will take. I like to create a simple
# timeline map first, highlighting points in time
# where each activity starts and stops.
#
# In our example, we will have 600 frames at 24FPS.
# Our animation will be split into 8 major sections
# of equal time duration. So that means each has
# 75 frames. Since 72 frames is an even 3 seconds,
# I use that instead, giving myself 24 frames of
# padding at the end in case I need them. That
# also provides a brief pause to let the viewer
# fully realize the final scene.
#
# The top Pyramid will move forward to the edge.
# Once its center (where its origin is at) is at
# the edge, it will stop and tilt. Once tilted
# to match the slope of the big pyramid, it will
# slide down the side of the pyramid until its
# front end hits the bottom. Now it will continue
# moving downward and forward, tilting back
# upright as it goes so the front edge never
# drops below the bottom of the ground. It will
# then continue forward until it collides with
# the "T" Object we created. Then the pyramid
# will stop moving, but the "T" will fall over
# forward.
#
#      FRAMES
#   START  END  ACTION
#   ----- ----- ----------------
#       1    72 Move forward to edge, no tilt
#      73   144 Full tilt
#     145   216 Drop 1 level
#     217   288 Drop 2 more levels
#     289   360 Front end hit ground, full tilt
#     361   432 Untilt and forward
#     433   504 Collision and start fall
#     505   576 Fall complete
#     577   600 Overflow
#
# Now we have a plan of attack, but always be
# flexible enough to tweak it as we go. Now
# we can coordinate our activity within each
# section to run in the frames allotted to it.
# It may sound very complex, but we will do it
# with only 21 keyframes. We can plot those
# out first, putting additional keyframes
# between those if we feel it is needed.
#
# Keyframes are tied directly to the Object.
# We need to specify the property (location,
# rotation, and/or scale) and frame number to
# be involved. Here are some examples:
#
# Location
#obj.keyframe_insert(data_path="location", frame=1)
# Rotation
#obj.keyframe_insert(data_path='rotation_euler', frame = 1)
# Scale
#obj.keyframe_insert(data_path='scale', frame = 1)
#
# If we had hundreds of keyframes to do, it would
# make sense to put the data in a list like:
#   kf = [
#       [object, location, loc_x, loc_y, loc_z, frame_#],
#       [object, rotation, rot_x, rot_y, rot_z, frame_#],
#       ---
#       ]
# Then a loop could process each line, setting the
# Object location (or rotation or scale) to the
# x,y,z values, and building the keyframe command
# to run.
#
# But our simple example has only 21 keyframes to
# be set, so we will just plot them outright.
# However, we will use the functions included with
# this script to build the keyframe command. So for
# each keyframe, we just need 2 lines of code, one
# to set the object location or rotation and one
# to call the proper function. Like this:
#
# Location
# Set
#obj.location = (x, y, z)
# Keyframe location
#kfl(obj, frame)
#
# or
#
# Rotation
# Set
#obj.rotation_euler = (x, y, z)
# Keyframe rotation
#kfr(obj, frame)
#
#
# Since we did not store a reference to any of
# the "Pyra.###" objects or the "My_T" object,
# we can do that now. We will only be using the
# "Pyra.229" (top pyramid) for our animation.
#
# Get "Pyra.229"
p229 = bpy.data.objects["Pyra.229"]
#
# Get "My_T"
myt = bpy.data.objects["My_T"]
#
#
#
# For Frame 1, we need to keyframe the initial
# values for every object which we animate. We
# do this to insure they do not start an action
# until told to do so in a later frame. We are
# only animating 2 objects so it is easy.
#
# Frame #1: Everything at start
#
# Pyra.229 location
p229.location = (0, 0, 1.2)
#
# Keyframe
kfl(p229, 1)
#
# Pyra.229 rotation
p229.rotation_euler = (0, 0, 0)
#
# Keyframe
kfr(p229, 1)
#
# My_T rotation
myt.rotation_euler = (0, 0, 0)
#
# Keyframe
kfr(myt, 1)
#
#
#
# Now we can just plot a keyframe for any change
# which occurs to the current action (if any) an
# object is doing.
#
#
# Frame #72: Pyra.229 reaches edge
#
# Pyra.229 location
p229.location = (0, -0.1, 1.2)
#
# Keyframe
kfl(p229, 72)
#
# Pyra.229 rotation
p229.rotation_euler = (0, 0, 0)
#
# Keyframe
kfr(p229, 72)
#
#
# Frame #108: Pyra.229 halfway tilted
#
# Pyra.229 location
p229.location = (0, -0.1, 1.2)
#
# Keyframe
kfl(p229, 108)
#
# Pyra.229 rotation
p229.rotation_euler = (0.5585, 0, 0)
#
# Keyframe
kfr(p229, 108)
#
#
# Frame #144: At edge, fully tilted
#
# Pyra.229 location
p229.location = (0, -0.1, 1.2)
#
# Keyframe
kfl(p229, 144)
#
# Pyra.229 rotation
p229.rotation_euler = (1.117, 0, 0)
#
# Keyframe
kfr(p229, 144)
#
#
# Frame #216: Pyra.229 dropped 1 level
#
# Pyra.229 location
p229.location = (0, -0.2, 1.0)
#
# Keyframe
kfl(p229, 216)
#
#
# Frame #288: Pyra.229 dropped 2 more levels
#
# Pyra.229 location
p229.location = (0, -0.4, 0.6)
#
# Keyframe
kfl(p229, 288)
#
#
# Frame #360: Pyra.229 front end at ground, fully tilted
#
# Pyra.229 location
p229.location = (0, -0.65, 0.1)
#
# Keyframe
kfl(p229, 360)
#
# Pyra.229 rotation
p229.rotation_euler = (1.117, 0, 0)
#
# Keyframe
kfr(p229, 360)
#
#
# Frame #396: Pyra.229 on ground, no tilt
#
# Pyra.229 location
p229.location = (0, -0.8, 0)
#
# Keyframe
kfl(p229, 396)
#
# Pyra.229 rotation
p229.rotation_euler = (0, 0, 0)
#
# Keyframe
kfr(p229, 396)
#
#
# Frame #432: Pyra.229 hits T and stops
#
# Pyra.229 location
p229.location = (0, -0.8492, 0)
#
# Keyframe
kfl(p229, 432)
#
#
# Frame #433: My_T upright, just before fall
#
# My_T rotation
myt.rotation_euler = (0, 0, 0)
#
# Keyframe
kfr(myt, 433)
#
#
# Frame #576: My_T fallen, stopped
#
# My_T rotation
myt.rotation_euler = (1.57, 0, 0)
#
# Keyframe
kfr(myt, 576)
#
#
# Just as we did for frame 1, we should set
# final keyframes for all of our animated
# objects on the last frame. This insures
# there will be no undesireable action
# between their former keyframe and the end
# of the video.
#
# Frame #600: Everything at finish
#
# Pyra.229 location
p229.location = (0, -0.8492, 0)
#
# Keyframe
kfl(p229, 600)
#
# Pyra.229 rotation
p229.rotation_euler = (0, 0, 0)
#
# Keyframe
kfr(p229, 600)
#
# My_T rotation
myt.rotation_euler = (1.57, 0, 0)
#
# Keyframe
kfr(myt, 600)
#
#
#
# There is a huge variety of ways to produce an
# animation with Blender. Above we just discussed
# a simple one, but the methods covered give us
# a solid foundation to build elaborate animations!
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
# Closing:
#
# In this tutorial, we covered many basic aspects
# of using Blenders Python API. Although there is
# a LOT more which can be done, I hope that this
# document will give beginners a great start in
# understanding the fundamental building blocks
# of using Blender via scripting.
#
#
# We will close this document in the same manner
# that we should finish all of our Blender Python
# scripts....
#
# Update the context to insure that all of the
# work we just did is shown properly!
#
# Update the view layer
bpy.context.view_layer.update()
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
#
# End of tutorial
#
#
